c  Although no one has reported any bugs in this program since it was
c  first used in 1988, be warned that there may still be some lurking.
c  If you find any, let me know.
c						K.R.J.  4 Aug. 1991

      subroutine quad ( a, b, c, r1r, r1i, r2r, r2i, nr )

c This routine finds the roots of the quadratic a*x**2 + b*x + c where a,
c b, c are real variables; r1r and r1i are the real and imaginary parts
c of one root; r2r and r2i are the real and imaginary parts of the other
c root; and nr is an integer flag set to
c      -2 if the input values are not normalized numbers
c         (r1r, r1i, r2r, r2i arbitrary.)
c      -1 if there are infinitely many roots (i.e., a=b=c=0).
c         (r1r, r1i, r2r, r2i arbitrary.)
c       0 if there are no roots (i.e., a=b=0, c=/=0).
c         (r1r, r1i, r2r, r2i arbitrary.)
c      11 if there is one root (i.e., a=0, b=/=0)
c         and it is a normalized number. 
c         (return the root in r1r, r1i with r2r, r2i arbitrary.)
c      10 if there is one root (i.e., a=0, b=/=0) and it is not
c         a normalized number. 
c         (r1r, r1i contain an approximate root which might be 0, 
c         a denormalized number, Inf or NaN; r2r, r2i arbitrary.)
c      22 if there are two roots (i.e., a=/=0) and both are
c         normalized numbers. 
c         (return one root in r1r, r1i and the other in r2r,r2i.)
c      21 if there are two roots (i.e., a=/=0) and only one is
c         a normalized number. 
c         (return the one normalized root in r1r, r1i and an approximate
c         root which might be 0, a denormalized number, Inf or NaN
c         in r2r, r2i.)
c      20 if there are two roots (i.e., a=/=0) and neither is
c         a normalized numbers.
c         (return approximations to the roots which might be 0, 
c         a denormalized number, Inf or NaN in r1r, r1i and r2r, r2i.)
c 
c This is a modified version of quad2.f that takes advantage of the IEEE
c floating pointing arithmetic on the Suns.  It is not quite as cautious
c as quad2.f in avoiding arithmetic exceptions, since these are better
c handled on the Suns than the Vax.  In this routine, we allow an
c exception to occur and then check for it later.  If a root is not
c sufficiently accurate or is outside the range of normalized numbers,
c then this is indicated through the flag nr and the computed
c approximation is returned.
c 
c I think (but I haven't spent enough time to verify it completely) that
c this routine satisfies both specification 1.1a and our modified version
c of specification 1.1b provided no arithmetic exceptions occur.  Any
c exceptions that do occur are flagged through nr.
c 
c It is written in single precision and uses double precision only to
c avoid loss of accuracy in computing the discriminant of the quadratic
c if the roots are close to being equal.  If this section were changed to
c single precision also, then the routine would still satisfy our
c modified version of specification 1.1b but not 1.1a.
c 
c This routine is machine dependent in that it assumes that the floating
c point numbers used conform to the IEEE standard in the sense that they
c are accurate, underflows are set to 0, overflows are set to +-Inf, and
c NaN and denormalized numbers can be detected.  However, if these last
c two groups don't exist, I believe that the routine will still function
c correctly.  The logical routines isInf, isNaN and isDnrm are used to
c detect +-Inf, all versions of NaN, and denormalized numbers:  isInf
c makes use of the the system supplied function flmax which returns the
c largest normalized machine number; isDnrm makes use of the the system
c supplied function flmin which returns the smallest positive normalized
c machine number; and isNaN makes use of the test x .ne. x which we
c assume is true for NaNs only.
c
c This subroutine also calls the intrinsic conversion functions real
c and dble as well as the subroutine switch supplied below which
c interchanges two real values.
c 					K.R.J.  20 Oct. 1988

      real a, b, c, abc, r1r, r1i, r2r, r2i, sdisc, f
      double precision disc
      integer nr
      logical isInf, isNaN, isDnrm

c First check for illegal input.

      if ( isInf(a) .or. isInf(b) .or. isInf(c) .or. 
     +     isNaN(a) .or. isNaN(b) .or. isNaN(c) .or. 
     +     isDnrm(a) .or. isDnrm(b) .or. isDnrm(c) ) then
        nr = -2
        return
      endif

c Next check for degenerate cases.

      if ( a .eq. 0. ) then

         if ( b .eq. 0. ) then

c           Both a and b .eq. 0

            if ( c .eq. 0. ) then
               nr = -1
            else
               nr = 0
            end if

            return

         else

c           Linear case -- a .eq. 0 but b .ne. 0.
c           Only one root exists; calculate it and check for 
c           arithmetic exceptions.

            r1r = -c / b
            r1i = 0.

            if ( ( r1r .eq. 0. .and. c .ne. 0. ) 
     +          .or. isDnrm(r1r) .or. isInf(r1r) ) then
               nr = 10
            else
               nr = 11
            endif

            return

         end if

      end if

c Non-degenerate case -- a .ne. 0.

c Treat special case c .eq. 0 first to avoid computational difficulties 
c below.

      if ( c .eq. 0. ) then

c        One root is zero; calculate the other root (r2r, r2i) and
c        check for arithmetic exceptions.

         r1r = 0.
         r1i = 0.
         r2r = -b / a
         r2i = 0.

         if ( ( r2r .eq. 0. .and. b .ne. 0. ) 
     +      .or. isDnrm(r2r) .or. isInf(r2r) ) then
           nr = 21
        else
           nr = 22
        endif

        return

      endif

c In remaining cases, both a .ne. 0 and c .ne. 0.
c First check if roots are close to being equal:  b**2 .approx. 4ac.

c Note that for IEEE floating point numbers, the following test works
c correctly even if there is an overflow;  no underflow can occur.
c The same holds for the other tests in this subroutine.

      if ( sign(1.,a)*sign(1.,c) .gt. 0. .and.
     +   1.9*sqrt(abs(a))*sqrt(abs(c)) .le. abs(b) .and.
     +   abs(b) .le. 2.1*sqrt(abs(a))*sqrt(abs(c)) ) then

c        The quadratic has "near double roots" in the sense that
c             0 .lt. 3.61*a*c .le. b**2 .le. 4.41*a*c.
c        Note that this implies that b .ne. 0, since a .ne. 0 and c .ne. 0.
c        Compute disc = b**2 - 4*a*c in double precision.
c        For IEEE floating point numbers, this cannot overflow or underflow.  
c        For other machines, it may be necessary to scale this computation.

         disc = dble(b)**2 - 4.d0*dble(a)*dble(c)

         if ( disc .ge. 0.d0 ) then

c           Real roots.  
c           Note that, from the requirement 
c              0 .lt. 3.61*a*c .le. b**2 .le. 4.41*a*c,
c           it follows that the true roots are nonzero.  
c           Consequently, if the computed roots are zero, then
c           a catastrophic underflow has occurred.

c           Also, note that because of the order in which the calculations
c           for r1r and r2r are performed, if an intermediate result 
c           overflows, the final result will be +-Inf, but no intermediate
c           loss of accuracy can result from an unnoticed intermediate 
c           denormalized number or underflow.  This remark applies to all
c           other computations below as well.

            sdisc = real(sqrt(disc))
            r1r = ( ( -b + sdisc ) / a ) / 2.
            r2r = ( ( -b - sdisc ) / a ) / 2.
            r1i = 0.
            r2i = 0.

            if ( r1r .eq. 0. .or. isDnrm(r1r) .or. isInf(r1r) ) then

               if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
                   nr = 20
               else
                  call switch(r1r,r2r)
                  nr = 21
               endif
             
            else

               if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
                  nr = 21
               else
                  nr = 22
               endif

            endif

            return

         else

c           Complex roots.  
c           Note that b .ne. 0 and sdisc .ne. 0.

            sdisc = real(sqrt(-disc))
            r1r = ( -b / a ) / 2.
            r2r = r1r
            r1i = ( sdisc / a ) / 2.
            r2i = - r1i

            if ( r1r .eq. 0. .or. isDnrm(r1r) .or. isInf(r1r) .or.
     +           r1i .eq. 0. .or. isDnrm(r1i) .or. isInf(r1i) ) 
     +      then
               nr = 20
            else
               nr = 22
            endif

            return

         endif

      endif

c End of double root case.

c Next consider case b**2 .ge. 4*abs(ac) .gt. 0.

      if ( abs(b) .ge. 2.*sqrt(abs(a))*sqrt(abs(c)) ) then

c        Note the following statement cannot overflow and an underflow
c        or an occurrence of a denormalized number is harmless.  
c        Also note that 1 .le. f .le. 1 + sqrt(2).

         f = 1. + sqrt(1. - sign(1.,a)*sign(1.,c)
     +            *4.*((sqrt(abs(a))*sqrt(abs(c)))/abs(b))**2 )

         r1r = -((f*b)/a)/2.
         r2r = -((2.*c)/b)/f
         r1i = 0.
         r2i = 0.

         if ( r1r .eq. 0. .or. isDnrm(r1r) .or. isInf(r1r) ) then

            if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
               nr = 20
            else
               call switch(r1r,r2r)
               nr = 21
            endif

         else

            if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
               nr = 21
            else
               nr = 22
            endif

         endif

         return

       endif

c The remaining case to consider is 0. .le. b**2 .lt. 4*abs(ac),
c which divides into 2 subcases depending upon the sign of ac.

      if ( sign(1.,a)*sign(1.,c) .lt. 0. ) then

c        Roots are real, b**2 .lt. 4*abs(ac) and ac .lt. 0.
c        Consequently, 0 .le. abc .le. 1 and 1 .le. f .le. 1 + sqrt(2).

         abc = (abs(b)/(sqrt(abs(a))*sqrt(abs(c))))/2.
         f = abc + sqrt(1. + abc**2)

         r1r = -sign(1.,a)*sign(1.,b)*(sqrt(abs(c))/sqrt(abs(a)))*f
         r2r = -sign(1.,c)*sign(1.,b)*(sqrt(abs(c))/sqrt(abs(a)))/f
         r1i = 0.
         r2i = 0.

         if ( ( abc .eq. 0. .and. b .ne. 0. ) .or. isDnrm(abc) ) then
            nr = 20
         else

            if ( r1r .eq. 0. .or. isDnrm(r1r) .or. isInf(r1r) ) then
               if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
                  nr = 20
               else
                  call switch(r1r,r2r)
                  nr = 21
               endif

            else

               if ( r2r .eq. 0. .or. isDnrm(r2r) .or. isInf(r2r) ) then
                  nr = 21
               else
                  nr = 22
               endif

            endif

         endif

         return

      else

c        Roots are complex, b**2 .lt. 4*(ac) and ac .gt. 0.

         r1r = -(b/a)/2.
         r2r = r1r
         r1i = (sqrt(abs(c))/sqrt(abs(a)))
     +         * sqrt(1. - (b/(sqrt(abs(a))*sqrt(abs(c)))/2.)**2)
         r2i = -r1i

         if ( ( r1r .eq. 0. .and. b .ne. 0. ) .or.
     +        isDnrm(r1r) .or. isInf(r1r) .or.
     +        r1i .eq. 0. .or. isDnrm(r1i) .or. isInf(r1i) ) then
            nr = 20
         else
            nr = 22
         endif

         return

      endif

      end



      logical function isInf(x)

c     This routine returns .true. if x is +-Inf and .false. otherwise.
c     It makes use of the system routine flmax which returns the largest
c     single precision floating point number.

      real x, flmax
      isInf = abs(x) .gt. flmax()

      return
      end


      logical function isNaN(x)

c     This routine returns .true. if x is NaN and .false. otherwise.
c     It assumes that x .ne. x is true for NaNs only.

      real x
      isNaN = x .ne. x

      return
      end


      logical function isDnrm(x)

c     This routine returns .true. if x is a denormalized floating point
c     number and .false. otherwise.
c     It makes use of the system routine flmin which returns the smallest
c     positive single precision floating point number.

      real x, flmin
      isDnrm = 0. .lt. abs(x) .and. abs(x) .lt. flmin()

      return
      end


      subroutine switch ( a, b )

c     This routine interchanges the values of a and b.

      real a, b, temp
      temp = a
      a = b
      b = temp

      return
      end
